/*
 * Win32 Device functions
 *
 * Copyright 2004 Codeweavers (Aric Stewart)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "config.h"
#include "wine/port.h"

#include <stdlib.h>
#include <stdio.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdarg.h>

#include "windef.h"
#include "winbase.h"
#include "winreg.h"
#include "winerror.h"
#include "winternl.h"
#include "winioctl.h"
#include "ntddstor.h"
#include "ntddscsi.h"
#include "kernel_private.h"
#include "wine/library.h"
#include "wine/unicode.h"
#include "wine/server.h"
#include "wine/debug.h"

/* linux includes */
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_SCSI_SG_H
#include <scsi/sg.h>
#endif
#ifdef HAVE_SCSI_SCSI_H
#include <scsi/scsi.h>
#endif
#ifdef HAVE_SCSI_SCSI_IOCTL_H
#include <scsi/scsi_ioctl.h>
#endif

WINE_DEFAULT_DEBUG_CHANNEL(device);


#define MAX_DEVICE_MODULES 32
#define SENSE_BUFF_LEN 32
#define INQUIRY_CMD 0x12
#define INQUIRY_CMDLEN 6
#define RESPONCE_LEN 96
#define DEF_TIMEOUT 60000


static CRITICAL_SECTION device_section;
static CRITICAL_SECTION_DEBUG critsect_debug =
{
    0, 0, &device_section,
    { &critsect_debug.ProcessLocksList, &critsect_debug.ProcessLocksList },
      0, 0, { 0, (DWORD)(__FILE__ ": device_section") }
};
static CRITICAL_SECTION device_section = { &critsect_debug, -1, 0, 0, 0, 0 };


/* scsi structs */

typedef struct scsi_idlun_ {
      int four_in_one;    /* 4 separate bytes of info
                             compacted into 1 int */
      int host_unique_id; /* distinguishes adapter cards from
                             same supplier */
} scsi_idlun;


/*
 * control what devices are openable 
 */
#define DEVICE_NAME_LEN 20
typedef WCHAR device_name[DEVICE_NAME_LEN+1];
static const device_name known_devices[] = {
    {'U','S','B','S','T','O','R',0},
    {'P','H','Y','S','I','C','A','L','D','R','I','V','E','1',0},
    {0}
    };

struct device_module
{
    device_name  device;
    HANDLE       handle;
};

static struct device_module device_modules[MAX_DEVICE_MODULES];


static int get_device_fd (HANDLE handle, DWORD access)
{
    int fd, ret;
    ret = wine_server_handle_to_fd( handle, access, &fd, NULL );
    if (ret) SetLastError( RtlNtStatusToDosError(ret) );
    return fd;
}

static inline void release_device_fd( HANDLE handle, int fd)
{
    wine_server_release_fd( handle, fd );
}


/* create a file handle to represent the device, Based on the VXD version */
static HANDLE open_device_handle( LPCWSTR name )
{
    const char *dir = wine_get_server_dir();
    int len;
    HANDLE ret;
    NTSTATUS status;
    OBJECT_ATTRIBUTES attr;
    UNICODE_STRING nameW;
    IO_STATUS_BLOCK io;
    DWORD l;

    l = strlenW(name);
    if (l > 8)
        l = 8;

    len = MultiByteToWideChar( CP_UNIXCP, 0, dir, -1, NULL, 0 );
    nameW.Length = ((len + l) * sizeof(WCHAR));
    nameW.MaximumLength = nameW.Length + sizeof(WCHAR);
    if (!(nameW.Buffer = HeapAlloc( GetProcessHeap(), 0, nameW.MaximumLength )))
    {
        SetLastError( ERROR_NOT_ENOUGH_MEMORY );
        return 0;
    }

    memset(nameW.Buffer, 0, nameW.MaximumLength);

    MultiByteToWideChar( CP_UNIXCP, 0, dir, -1, nameW.Buffer, len );
    nameW.Buffer[len-1] = '/';
    memcpy( nameW.Buffer + len, name, l * sizeof(WCHAR) );
    nameW.Buffer[len + l] = 0;

    TRACE("file is %s\n",debugstr_w(nameW.Buffer));
    attr.Length = sizeof(attr);
    attr.RootDirectory = 0;
    attr.Attributes = 0;
    attr.ObjectName = &nameW;
    attr.SecurityDescriptor = NULL;
    attr.SecurityQualityOfService = NULL;

    memset(&io,0,sizeof(io));

    status = NtCreateFile( &ret, 0, &attr, &io, NULL, 0,
                           FILE_SHARE_READ|FILE_SHARE_WRITE, FILE_OPEN_IF,
                           FILE_SYNCHRONOUS_IO_ALERT, NULL, 0 );
    if (status)
    {
        ret = 0;
        SetLastError( RtlNtStatusToDosError(status) );
    }
    RtlFreeUnicodeString( &nameW );
    return ret;
}

/* load a Device and return a file handle to it */
HANDLE Device_Open( LPCWSTR filenameW, DWORD access, SECURITY_ATTRIBUTES *sa )
{
    int i;
    HANDLE handle;
    device_name Name; 
    LPWSTR ptr;

    /* get the device name */
    ptr = strchrW(filenameW,'#'); 
    if (ptr)
        i = min(DEVICE_NAME_LEN,ptr - filenameW);
    else
        i = min(DEVICE_NAME_LEN,strlenW(filenameW));

    memcpy(Name, filenameW, i * sizeof(WCHAR));
    Name[i] = 0;

    TRACE("Trying to open device %s\n",debugstr_w(Name));
    
    /* check the device name agienst handled devices */
    for (i = 0; known_devices[i][0]; i++)
    {
        if (strcmpiW(Name,known_devices[i])==0)
            break;
    }
    if (known_devices[i][0]==0)
    {
        FIXME("Unhandled device %s\n",debugstr_w(Name));
        SetLastError(ERROR_FILE_NOT_FOUND);
        return 0;
    }

    RtlEnterCriticalSection( &device_section );

    for (i = 0; i < MAX_DEVICE_MODULES; i++)
    {
        if (strcmpiW(device_modules[i].device,Name) == 0)
        {
            handle = device_modules[i].handle;
            goto done;  /* already registered */
        }
        if (!device_modules[i].handle)  /* new one, register it */
        {
            int fd;

            /* get a file handle to the dummy file */
            if (!(handle = open_device_handle( Name )))
            {
                goto done;
            }
            wine_server_handle_to_fd( handle, 0, &fd, NULL );
            strcpyW(device_modules[i].device,Name);
            device_modules[i].handle = handle;
            wine_server_release_fd( handle, fd );
            goto done;
        }
    }

    ERR("too many open device modules, please report\n" );
    CloseHandle( handle );
    handle = 0;

done:
    RtlLeaveCriticalSection( &device_section );
    if (!DuplicateHandle( GetCurrentProcess(), handle, GetCurrentProcess(), &handle, 0,
                          (sa && (sa->nLength>=sizeof(*sa)) && sa->bInheritHandle),
                          DUP_HANDLE_SAME_ACCESS ))
        handle = 0;
    if (!handle)
        SetLastError(ERROR_FILE_NOT_FOUND);
    return handle;
}

static int get_scsi_generic_from_block(HANDLE hDevice)
{
#ifdef SCSI_IOCTL_GET_IDLUN
    scsi_idlun  idlun,idlun2;
    int i = 0;
    int error_count = 0;
    int fd;
    int rc;
    int sg_device = -1;

    fd = get_device_fd(hDevice, 0);

    memset(&idlun, 0, sizeof(idlun));
    rc = ioctl(fd,SCSI_IOCTL_GET_IDLUN,&idlun);

    release_device_fd(hDevice,fd);

    if (rc != 0)
    {
        TRACE("SCSI_IOCTL_GET_IDLUN failed\n");
        return -1; /* not a scsi device */
    }

    TRACE("finding device with idlun = { 0x%x, 0x%x }\n", idlun.four_in_one, idlun.host_unique_id);

    /* find the generic scsi device with the same scsi id */

    while (error_count < 12)
    {
        char name[60];

        sprintf(name,"/dev/sg%i",i);
        TRACE("Trying device %s\n",name);
        
        sg_device = open(name,O_RDONLY|O_NONBLOCK);
        if (sg_device != -1)
        {
            memset(&idlun2,0,sizeof(idlun2));
            ioctl(sg_device,SCSI_IOCTL_GET_IDLUN,&idlun2);
            if ((idlun2.four_in_one == idlun.four_in_one) &&
                (idlun2.host_unique_id == idlun.host_unique_id))
                break;
            else {
                TRACE("opened %s, but it didn't match\n", name);
                close(sg_device);
                sg_device = -1;
            }
        }
        else
            error_count ++;
        i++;
    }

    return sg_device;
#else
    FIXME("SCSI_IOCTL_GET_IDLUN not supported\n");
    return -1;
#endif
}

static BOOL ioctl_scsi_pass_through(HANDLE hDevice, SCSI_PASS_THROUGH_DIRECT *spt, LPDWORD lpcbBytesReturned)
{
    UCHAR i;
    int sg_device;
    int rc = -1;
#ifdef HAVE_SG_IO_HDR_T_INTERFACE_ID
    sg_io_hdr_t io_hdr;
#endif
    HANDLE block_device;

    /* FIXME: big hack */
    block_device = CreateFileA("\\\\.\\D:", GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
    if (block_device == INVALID_HANDLE_VALUE)
    {
        ERR("Couldn't open block device, error %d\n", GetLastError());
        return FALSE;
    }
    /* end big hack */

    sg_device = get_scsi_generic_from_block(block_device);

    CloseHandle(block_device);
    
    if (sg_device == -1)
    {
        TRACE("No matching openable scsi generic device found\n");
        SetLastError(ERROR_INVALID_FUNCTION);
        return FALSE;
    }

    TRACE("Cdb:");
    for (i = 0; i < spt->CdbLength; i++)
        TRACE(" %02x", spt->Cdb[i]);
    TRACE("\n");

    TRACE("spt->DataIn = %d\n", spt->DataIn);
    TRACE("spt->DataTransferLength = %d\n", spt->DataTransferLength);
    TRACE("spt->SenseInfoLength = %d\n", spt->SenseInfoLength);
    TRACE("spt->SenseInfoOffset = 0x%x\n", spt->SenseInfoOffset);

#ifdef HAVE_SG_IO_HDR_T_INTERFACE_ID
    memset(&io_hdr,0,sizeof(io_hdr));

    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = spt->CdbLength;
    io_hdr.mx_sb_len = spt->SenseInfoLength;
    if (spt->DataIn == SCSI_IOCTL_DATA_IN)
        io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
    else if (spt->DataIn == SCSI_IOCTL_DATA_OUT)
        io_hdr.dxfer_direction = SG_DXFER_TO_DEV;
#if 0
    else
        io_hdr.dxfer_direction = SG_DXFER_UNKNOWN;
#else
    else
        FIXME("spt_DataIn = %d\n", spt->DataIn);
#endif
    io_hdr.dxfer_len = spt->DataTransferLength;
    io_hdr.dxferp = spt->DataBuffer;
    io_hdr.cmdp = spt->Cdb;
    io_hdr.sbp = (BYTE *)spt + spt->SenseInfoOffset;
    io_hdr.timeout = DEF_TIMEOUT; /* FIXME: use sbt->TimeOutValue */

    rc = ioctl(sg_device, SG_IO, &io_hdr);
    spt->ScsiStatus = 0; /* FIXME: use rc */
    *lpcbBytesReturned = io_hdr.dxfer_len;
#endif /* HAVE_SG_IO_HDR_T_INTERFACE_ID */

    close(sg_device);
    if (rc < 0)
    {
        ERR("ioctl SG_IO failed with error %d\n", rc);
        SetLastError(ERROR_INVALID_FUNCTION);
        return FALSE;
    }

    return TRUE;
}

/*
 * This function is very incomplete. It totally is ONLY targeting the ipod
 * device at this time.
 */

BOOL Device_DeviceIoControl(HANDLE hDevice, DWORD dwIoControlCode,
                            LPVOID lpvInBuffer, DWORD cbInBuffer,
                            LPVOID lpvOutBuffer, DWORD cbOutBuffer,
                            LPDWORD lpcbBytesReturned,
                            LPOVERLAPPED lpOverlapped)
{
    TRACE( "(%p,%x,%p,%d,%p,%d,%p,%p)\n",
           hDevice,dwIoControlCode,lpvInBuffer,cbInBuffer,
           lpvOutBuffer,cbOutBuffer,lpcbBytesReturned,lpOverlapped );

    switch (dwIoControlCode)
    {
    case IOCTL_STORAGE_GET_DEVICE_NUMBER:
    {
        /* 
         * Bad Codeweavers, hacking this function so that is simple returns some
         * data instead of truly calling the ioctl functions to retreave the
         * data. I will fix this soon. 
         */
        PSTORAGE_DEVICE_NUMBER dev = (PSTORAGE_DEVICE_NUMBER)lpvOutBuffer;

        TRACE("IOCTL_STORAGE_GET_DEVICE_NUMBER\n");
       
        dev->DeviceType = FILE_DEVICE_DISK;
        dev->DeviceNumber = 1;
        dev->PartitionNumber = 0; 
        *lpcbBytesReturned = sizeof(STORAGE_DEVICE_NUMBER);
        return TRUE;
    }
    case IOCTL_STORAGE_QUERY_PROPERTY:
    {
        /* 
         * Bad codeweavers part II,  yes since i am really only intrested in 
         * iPods right now, this will only work on usb devices. Because USB
         * storage is based on scsi i can use the scsi ioctl functions to get
         * the information i need
         */
        PSTORAGE_PROPERTY_QUERY query = (PSTORAGE_PROPERTY_QUERY)lpvInBuffer;
        PSTORAGE_DEVICE_DESCRIPTOR output =
                                    (PSTORAGE_DEVICE_DESCRIPTOR)lpvOutBuffer;
        int rc = -1;
        int sg_device = -1;
#ifdef HAVE_SG_IO_HDR_T_INTERFACE_ID
        sg_io_hdr_t io_hdr;
        unsigned char sense[SENSE_BUFF_LEN];
#endif
        char responce[RESPONCE_LEN];
        unsigned char CommandBlock[INQUIRY_CMDLEN] = {INQUIRY_CMD, 0, 0, 0,
                                                      0, 0};


        TRACE("IOCTL_STORAGE_QUERY_PROPERTY\n");

        TRACE("PropertyId = 0x%x, QueryType = 0x%x\n",query->PropertyId, query->QueryType);

        sg_device = get_scsi_generic_from_block(hDevice);

        if (sg_device == -1)
        {
            TRACE("No matching openable scsi generic device found\n");
            SetLastError(ERROR_INVALID_FUNCTION);
            return FALSE;
        }

        /* Prepare for the inquery call*/

        CommandBlock[4] = (unsigned char)RESPONCE_LEN;
        memset(responce,0,sizeof(responce));
#ifdef HAVE_SG_IO_HDR_T_INTERFACE_ID
        memset(&io_hdr,0,sizeof(io_hdr));

        io_hdr.interface_id = 'S';
        io_hdr.cmd_len = sizeof (CommandBlock);
        io_hdr.mx_sb_len = sizeof(sense);
        io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
        io_hdr.dxfer_len = sizeof(responce);
        io_hdr.dxferp = responce;
        io_hdr.cmdp = CommandBlock;
        io_hdr.sbp = sense;
        io_hdr.timeout = DEF_TIMEOUT;
       
        rc = ioctl(sg_device,SG_IO,&io_hdr);
#endif

        close(sg_device);
        if (rc)
        {
            ERR("ioctl SG_IO failed\n");
            SetLastError(ERROR_INVALID_FUNCTION);
            return FALSE;
        }

        memset(output,0,sizeof(output)+RESPONCE_LEN);
        output->Version =sizeof(output);
        output->Size = sizeof(output) + sizeof(responce);
        output->BusType = BusTypeUsb;
        output->RawPropertiesLength = sizeof(responce);
        memcpy(&(output->RawDeviceProperties),responce,sizeof(responce));
        *lpcbBytesReturned = sizeof(STORAGE_DEVICE_DESCRIPTOR)+sizeof(responce);

        return TRUE;
    }
    case IOCTL_SCSI_PASS_THROUGH:
    {
        SCSI_PASS_THROUGH *spt = (SCSI_PASS_THROUGH *)lpvInBuffer;
        ULONG_PTR DataBufferOffset = spt->DataBufferOffset;
        BOOL ret;

        TRACE("IOCTL_SCSI_PASS_THROUGH\n");

        /* convert SCSI_PASS_THROUGH -> SCSI_PASS_THROUGH_DIRECT */
        spt->DataBufferOffset = (ULONG_PTR)((char *)spt + spt->DataBufferOffset);

        ret = ioctl_scsi_pass_through(hDevice, (SCSI_PASS_THROUGH_DIRECT *)spt, lpcbBytesReturned);

        spt->DataBufferOffset = DataBufferOffset;
        return ret;
    }
    case IOCTL_SCSI_PASS_THROUGH_DIRECT:
        TRACE("IOCTL_SCSI_PASS_THROUGH_DIRECT\n");
        return ioctl_scsi_pass_through(hDevice, lpvInBuffer, lpcbBytesReturned);

    default:
        return FALSE;  /* fall back to normal handling */
    }
}
